import zipfile
import requests
from io import BytesIO
import rasterio
from rasterio.io import MemoryFile
import pandas as pd
import geopandas as gpd
import numpy as np

from .utils import get_population_data_links, get_available_pop_countries


def get_population_data(country: str, use_tif = False):
    """Extract Meta High Resolution Population Density dataset from HDX based on country name. 
    Due to raster band sparsity for recent time periods, prioritises construction of indicators 
    from .csv file over .tif. Values are rounded to integers for ease of intepretation and for 
    convenient raster plotting.

    Args:
        country (str): Country name

    Returns:
        pd.DataFrame/np.ndarray: Dataframe (csv) or array (tif) with population information
    """    
    
    general_pop_dict = get_population_data_links()

    try:
        general_pop_dict[country]
    except KeyError:
        print(f"Population data is not available for {country}. Please use utils.get_available_pop_countries function to view which countries are available.")

    groups = ['pop', 'men', 'women', 'elderly','youth','children']

    if not use_tif:
        groups_df = []
        target_cols = []
        for group in groups:
            data_link = general_pop_dict[country][f'{group}_csv_url']
            df = pd.read_csv(data_link)
            for colname in list(df.columns):
                if 'population' in colname:
                    target_col = colname
            df[target_col] = df[target_col].astype(int)
            pop_gdf = gpd.GeoDataFrame(
                data=df[target_col], 
                crs = 'epsg:4326',
                geometry = gpd.points_from_xy(df.longitude, df.latitude))
            groups_df.append(pop_gdf)
            target_cols.append(target_col)

        return groups_df, target_cols

    src_list = []
    rc_list = []
    for group in groups:
        data_link = general_pop_dict[country][f'{group}_tif_url']
        filename = data_link.split('/')[-1][:-12]
        req = requests.get(data_link)

        try:
        # Extract and read tif into memory from zip file
            with zipfile.ZipFile(BytesIO(req.content)) as myzip:
                with myzip.open(f'{filename}.tif', 'r') as myfile:
                    pop_map = myfile.read()
        except KeyError:
            filename = filename.replace(filename[:3], filename[:3].upper())
            with zipfile.ZipFile(BytesIO(req.content)) as myzip:
                with myzip.open(f'{filename}.tif', 'r') as myfile:
                    pop_map = myfile.read()

        # Read TIF in-memory
        with MemoryFile(pop_map) as memfile:
            src = memfile.open()
            rc = src.read(out_dtype='uint16')   

        # Update TIF in-memory
        meta = src.meta
        meta.update({
        "dtype": 'uint16',
            "nodata": 0
        })

        with MemoryFile() as myfile:
            src = myfile.open(** meta)
            src.write(rc)
            rc = src.read()
        
        src_list.append(src)
        rc_list.append(rc)
    
    return src_list, rc_list



# Geometric helper functions
from shapely.ops import split
from functools import partial
from shapely.geometry import Point, Polygon, LineString, MultiPolygon
import networkx as nx
import geopandas as gpd
import numpy as np

def project_gdf(gdf):
    mean_longitude = gdf["geometry"].representative_point().x.mean()

    # Compute UTM crs
    utm_zone = int(np.floor((mean_longitude + 180) / 6) + 1)
    utm_crs = f"+proj=utm +zone={utm_zone} +ellps=WGS84 +datum=WGS84 +units=m +no_defs"

    # project the GeoDataFrame to the UTM CRS
    gdf_proj = gdf.to_crs(utm_crs)
    # print(f"Projected to {gdf_proj.crs}")
    
    return gdf_proj

def buffer_polygon(gdf, bandwidth = 200):
    buffer_zone = project_gdf(gdf).buffer(bandwidth)
    buffered_gdf = buffer_zone.to_crs(4326)
    return buffered_gdf

def great_circle_vec(lat1, lng1, lat2, lng2, earth_radius=6_371_009):
    """
    Calculate great-circle distances between pairs of points.

    Vectorized function to calculate the great-circle distance between two
    points' coordinates or between arrays of points' coordinates using the
    haversine formula. Expects coordinates in decimal degrees.

    Parameters
    ----------
    lat1 : float or numpy.array of float
        first point's latitude coordinate
    lng1 : float or numpy.array of float
        first point's longitude coordinate
    lat2 : float or numpy.array of float
        second point's latitude coordinate
    lng2 : float or numpy.array of float
        second point's longitude coordinate
    earth_radius : float
        earth's radius in units in which distance will be returned (default is
        meters)

    Returns
    -------
    dist : float or numpy.array of float
        distance from each (lat1, lng1) to each (lat2, lng2) in units of
        earth_radius
    """
    y1 = np.deg2rad(lat1)
    y2 = np.deg2rad(lat2)
    dy = y2 - y1

    x1 = np.deg2rad(lng1)
    x2 = np.deg2rad(lng2)
    dx = x2 - x1

    h = np.sin(dy / 2) ** 2 + np.cos(y1) * np.cos(y2) * np.sin(dx / 2) ** 2
    h = np.minimum(1, h)  # protect against floating point errors
    arc = 2 * np.arcsin(np.sqrt(h))

    # return distance in units of earth_radius
    return arc * earth_radius


def add_edge_lengths(G, precision=3):
    uvk = tuple(G.edges)
    x = G.nodes(data="x")
    y = G.nodes(data="y")
    
    try:
        # two-dimensional array of coordinates: y0, x0, y1, x1
        c = np.array([(y[u], x[u], y[v], x[v]) for u, v, k in uvk])
        
    except KeyError:  # pragma: no cover
        raise KeyError("some edges missing nodes, possibly due to input data clipping issue")
        
    dists = great_circle_vec(c[:, 0], c[:, 1], c[:, 2], c[:, 3]).round(precision)
    dists[np.isnan(dists)] = 0
    nx.set_edge_attributes(G, values=dict(zip(uvk, dists)), name="length")
    
    return G

def _is_endpoint(G, node, strict=True):
    """
    Is node a true endpoint of an edge.

    Return True if the node is a "real" endpoint of an edge in the network,
    otherwise False. OSM data includes lots of nodes that exist only as points
    to help streets bend around curves. An end point is a node that either:
    1) is its own neighbor, ie, it self-loops.
    2) or, has no incoming edges or no outgoing edges, ie, all its incident
    edges point inward or all its incident edges point outward.
    3) or, it does not have exactly two neighbors and degree of 2 or 4.
    4) or, if strict mode is false, if its edges have different OSM IDs.

    Parameters
    ----------
    G : networkx.MultiDiGraph
        input graph
    node : int
        the node to examine
    strict : bool
        if False, allow nodes to be end points even if they fail all other rules
        but have edges with different OSM IDs

    Returns
    -------
    bool
    """
    neighbors = set(list(G.predecessors(node)) + list(G.successors(node)))
    n = len(neighbors)
    d = G.degree(node)

    # rule 1
    if node in neighbors:
        # if the node appears in its list of neighbors, it self-loops
        # this is always an endpoint.
        return True

    # rule 2
    elif G.out_degree(node) == 0 or G.in_degree(node) == 0:
        # if node has no incoming edges or no outgoing edges, it is an endpoint
        return True

    # rule 3
    elif not (n == 2 and (d == 2 or d == 4)):
        # else, if it does NOT have 2 neighbors AND either 2 or 4 directed
        # edges, it is an endpoint. either it has 1 or 3+ neighbors, in which
        # case it is a dead-end or an intersection of multiple streets or it has
        # 2 neighbors but 3 degree (indicating a change from oneway to twoway)
        # or more than 4 degree (indicating a parallel edge) and thus is an
        # endpoint
        return True

    # rule 4
    elif not strict:
        # non-strict mode: do its incident edges have different OSM IDs?
        osmids = []

        # add all the edge OSM IDs for incoming edges
        for u in G.predecessors(node):
            for key in G[u][node]:
                osmids.append(G.edges[u, node, key]["osmid"])

        # add all the edge OSM IDs for outgoing edges
        for v in G.successors(node):
            for key in G[node][v]:
                osmids.append(G.edges[node, v, key]["osmid"])

        # if there is more than 1 OSM ID in the list of edge OSM IDs then it is
        # an endpoint, if not, it isn't
        return len(set(osmids)) > 1

    # if none of the preceding rules returned true, then it is not an endpoint
    else:
        return False
    
def _build_path(G, endpoint, endpoint_successor, endpoints):
    """
    Build a path of nodes from one endpoint node to next endpoint node.

    Parameters
    ----------
    G : networkx.MultiDiGraph
        input graph
    endpoint : int
        the endpoint node from which to start the path
    endpoint_successor : int
        the successor of endpoint through which the path to the next endpoint
        will be built
    endpoints : set
        the set of all nodes in the graph that are endpoints

    Returns
    -------
    path : list
        the first and last items in the resulting path list are endpoint
        nodes, and all other items are interstitial nodes that can be removed
        subsequently
    """
    # start building path from endpoint node through its successor
    path = [endpoint, endpoint_successor]

    # for each successor of the endpoint's successor
    for successor in G.successors(endpoint_successor):
        if successor not in path:
            # if this successor is already in the path, ignore it, otherwise add
            # it to the path
            path.append(successor)
            while successor not in endpoints:
                # find successors (of current successor) not in path
                successors = [n for n in G.successors(successor) if n not in path]

                # 99%+ of the time there will be only 1 successor: add to path
                if len(successors) == 1:
                    successor = successors[0]
                    path.append(successor)

                # handle relatively rare cases or OSM digitization quirks
                elif len(successors) == 0:
                    if endpoint in G.successors(successor):
                        # we have come to the end of a self-looping edge, so
                        # add first node to end of path to close it and return
                        return path + [endpoint]
                    else:  # pragma: no cover
                        # this can happen due to OSM digitization error where
                        # a one-way street turns into a two-way here, but
                        # duplicate incoming one-way edges are present
                        utils.log(
                            f"Unexpected simplify pattern handled near {successor}", level=lg.WARN
                        )
                        return path
                else:  # pragma: no cover
                    # if successor has >1 successors, then successor must have
                    # been an endpoint because you can go in 2 new directions.
                    # this should never occur in practice
                    raise Exception(f"Unexpected simplify pattern failed near {successor}")

            # if this successor is an endpoint, we've completed the path
            return path

    # if endpoint_successor has no successors not already in the path, return
    # the current path: this is usually due to a digitization quirk on OSM
    return path

def _get_paths_to_simplify(G, strict=True):
    """
    Generate all the paths to be simplified between endpoint nodes.

    The path is ordered from the first endpoint, through the interstitial nodes,
    to the second endpoint.

    Parameters
    ----------
    G : networkx.MultiDiGraph
        input graph
    strict : bool
        if False, allow nodes to be end points even if they fail all other rules
        but have edges with different OSM IDs

    Yields
    ------
    path_to_simplify : list
    """
    # first identify all the nodes that are endpoints
    endpoints = set([n for n in G.nodes if _is_endpoint(G, n, strict=strict)])

    # for each endpoint node, look at each of its successor nodes
    for endpoint in endpoints:
        for successor in G.successors(endpoint):
            if successor not in endpoints:
                # if endpoint node's successor is not an endpoint, build path
                # from the endpoint node, through the successor, and on to the
                # next endpoint node
                yield _build_path(G, endpoint, successor, endpoints)
                
def simplify_graph(G, strict=True, remove_rings=True):
    """
    Simplify a graph's topology by removing interstitial nodes.

    Simplifies graph topology by removing all nodes that are not intersections
    or dead-ends. Create an edge directly between the end points that
    encapsulate them, but retain the geometry of the original edges, saved as
    a new `geometry` attribute on the new edge. Note that only simplified
    edges receive a `geometry` attribute. Some of the resulting consolidated
    edges may comprise multiple OSM ways, and if so, their multiple attribute
    values are stored as a list.

    Parameters
    ----------
    G : networkx.MultiDiGraph
        input graph
    strict : bool
        if False, allow nodes to be end points even if they fail all other
        rules but have incident edges with different OSM IDs. Lets you keep
        nodes at elbow two-way intersections, but sometimes individual blocks
        have multiple OSM IDs within them too.
    remove_rings : bool
        if True, remove isolated self-contained rings that have no endpoints

    Returns
    -------
    G : networkx.MultiDiGraph
        topologically simplified graph, with a new `geometry` attribute on
        each simplified edge
    """
    if "simplified" in G.graph and G.graph["simplified"]:  # pragma: no cover
        raise Exception("This graph has already been simplified, cannot simplify it again.")


    # define edge segment attributes to sum upon edge simplification
    attrs_to_sum = {"length"}

    # make a copy to not mutate original graph object caller passed in
    G = G.copy()
    initial_node_count = len(G)
    initial_edge_count = len(G.edges)
    all_nodes_to_remove = []
    all_edges_to_add = []

    # generate each path that needs to be simplified
    for path in _get_paths_to_simplify(G, strict=strict):

        # add the interstitial edges we're removing to a list so we can retain
        # their spatial geometry
        path_attributes = dict()
        for u, v in zip(path[:-1], path[1:]):

            # there should rarely be multiple edges between interstitial nodes
            # usually happens if OSM has duplicate ways digitized for just one
            # street... we will keep only one of the edges (see below)
            edge_count = G.number_of_edges(u, v)
            if edge_count != 1:
                utils.log(f"Found {edge_count} edges between {u} and {v} when simplifying")

            # get edge between these nodes: if multiple edges exist between
            # them (see above), we retain only one in the simplified graph
            edge_data = G.edges[u, v, 0]
            for attr in edge_data:
                if attr in path_attributes:
                    # if this key already exists in the dict, append it to the
                    # value list
                    path_attributes[attr].append(edge_data[attr])
                else:
                    # if this key doesn't already exist, set the value to a list
                    # containing the one value
                    path_attributes[attr] = [edge_data[attr]]

        # consolidate the path's edge segments' attribute values
        for attr in path_attributes:
            if attr in attrs_to_sum:
                # if this attribute must be summed, sum it now
                path_attributes[attr] = sum(path_attributes[attr])
            elif len(path_attributes[attr]) == 1:
                # if there's only 1 unique value in this attribute list,
                # consolidate it to the single value (the zero-th):
                path_attributes[attr] = path_attributes[attr][0]
            else:
                # otherwise, if there are multiple values, keep one of each
                path_attributes[attr] = tuple(path_attributes[attr])

        # construct the new consolidated edge's geometry for this path
        path_attributes["geometry"] = LineString(
            [Point((G.nodes[node]["x"], G.nodes[node]["y"])) for node in path]
        )

        # add the nodes and edge to their lists for processing at the end
        all_nodes_to_remove.extend(path[1:-1])
        all_edges_to_add.append(
            {"origin": path[0], "destination": path[-1], "attr_dict": path_attributes}
        )

    # for each edge to add in the list we assembled, create a new edge between
    # the origin and destination
    for edge in all_edges_to_add:
        G.add_edge(edge["origin"], edge["destination"], **edge["attr_dict"])

    # finally remove all the interstitial nodes between the new edges
    G.remove_nodes_from(set(all_nodes_to_remove))

    if remove_rings:
        # remove any connected components that form a self-contained ring
        # without any endpoints
        wccs = nx.weakly_connected_components(G)
        nodes_in_rings = set()
        for wcc in wccs:
            if not any(_is_endpoint(G, n) for n in wcc):
                nodes_in_rings.update(wcc)
        G.remove_nodes_from(nodes_in_rings)

    # mark graph as having been simplified
    G.graph["simplified"] = True

    return G

def graph_to_gdf(G, nodes=False, edges=False, dual=False):

    crs = G.graph['crs']
    if nodes:
        if not G.nodes:  # pragma: no cover
            raise ValueError("graph contains no nodes")
            
        nodes, data = zip(*G.nodes(data=True))
        
        # convert node x/y attributes to Points for geometry column
        geom = (Point(d["x"], d["y"]) for d in data)
        gdf_nodes = gpd.GeoDataFrame(data, index=nodes, crs=crs, geometry=list(geom))
         
        if not dual:
            gdf_nodes.index.rename("osmid", inplace=True)
        
    if edges:
        if not G.edges:  # pragma: no cover
            raise ValueError("graph contains no edges")
        
        if not dual: 
            u, v, k, data = zip(*G.edges(keys=True, data=True))
        else: 
            u, v, data = zip(*G.edges(data=True))
            
        x_lookup = nx.get_node_attributes(G, "x")
        y_lookup = nx.get_node_attributes(G, "y")
        def make_geom(u, v, data, x= x_lookup, y= y_lookup):
            if "geometry" in data:
                return data["geometry"]
            else:
                return LineString((Point((x[u], y[u])), Point((x[v], y[v]))))
            
        geom = map(make_geom, u, v, data)
        gdf_edges = gpd.GeoDataFrame(data, crs=crs, geometry=list(geom))


        # add u, v, key attributes as index
        gdf_edges["u"] = u
        gdf_edges["v"] = v
        if not dual: 
            gdf_edges["key"] = k
            gdf_edges.set_index(["u", "v", "key"], inplace=True)
        else: 
            gdf_edges.set_index(["u", "v"], inplace=True)
    if nodes and edges:
        return gdf_nodes, gdf_edges
    elif nodes:
        return gdf_nodes
    elif edges:
        return gdf_edges

def fill_and_expand(gdf):
    """Function to expand multilinestring and multipolygons into Polygon,
     to allow for overlay operation when computing building attributes.

    Args:
        gdf (geopandas.geodataframe.GeoDataFrame): Dataframe consisting of network nodes.

    Returns:
        geopandas.geodataframe.GeoDataFrame: Modified dataframe with homogenous geometry type. 
    """    

    for i, geom in enumerate(gdf.geometry):

        if geom.geom_type == 'LineString':
            gdf = gdf.drop(i, axis=0)

        elif geom.geom_type=='MultiLineString':
            linestring = gdf[gdf.index == i]
            linestring_exploded = linestring.explode(index_parts=True)

            # Find bounding polygon
            poly_len = len(linestring_exploded)
            list_of_minx = list(linestring_exploded.bounds.minx)
            min_ind = list_of_minx.index(min(list_of_minx))
            
            # Create list of hole polygons
            holes = []
            for k, geom in enumerate(linestring_exploded.geometry):
                if (k != min_ind) and (geom.geom_type!='LineString'):
                    holes.append(list(geom.coords))

            try:
                polygon_geom = Polygon(list(list(linestring_exploded.geometry.iloc[min_ind].coords)), 
                                  holes = holes)
                polygon = gpd.GeoDataFrame(index=[0], crs=linestring.crs, geometry=[polygon_geom]) 
                gdf = gdf.drop(i, axis=0)
                gdf = gpd.GeoDataFrame(pd.concat([gpd, polygon], ignore_index=True), crs=gpd.crs) 

            except ValueError:
                gdf = gdf.drop(i, axis=0)          

    # Expand MultiPolygons
    gdf = gdf.explode(index_parts=True)
    
    return gdf


# Map class utility functions
from __future__ import annotations
import os
import json
import pkg_resources
from typing import Optional
from IPython.display import display
from ipyleaflet import DrawControl


def get_country_centroids():
    data_path = pkg_resources.resource_filename('urbanity', "map_data/country.json")
    with open(data_path) as f:
        country_dict = json.load(f)

    return country_dict

def get_population_data_links():
    data_path = pkg_resources.resource_filename('urbanity', "map_data/links_general.json")
    with open(data_path) as f:
        general_pop_dict = json.load(f)
    return general_pop_dict

def get_available_pop_countries():
    general_pop_dict = set(get_population_data_links())
    print(sorted(general_pop_dict))

def get_available_countries():
    country_dict = set(get_country_centroids())
    print(sorted(country_dict))

def gdf_to_tensor():
    pass

def finetune_poi(df, target, relabel_dict, n=30):
    
    df2 = df.copy()
    for k,v in relabel_dict.items():
        df2[target] = df2[target].replace(k, v)
    
    # remove categories with less than n instances
    
    cat_list = df2[target].value_counts().index
    cat_mask = (df2[target].value_counts() > n).values
    selected = set(cat_list[cat_mask])
    
    df2 = df2[df2[target].isin(selected)]
    
    return df2



# main module
import os
import json
import time
import math
from unittest import skip
import pkg_resources
from typing import Optional, Union
from webbrowser import get

from .utils import get_country_centroids, finetune_poi
from .geom import *
from .population import get_population_data

import networkx as nx
from networkx import MultiDiGraph
import pandas as pd
import geopandas as gpd
from shapely.geometry import Polygon
import ipyleaflet
from ipyleaflet import basemaps, basemap_to_tiles, Icon, Marker, LayersControl, LayerGroup, DrawControl, FullScreenControl, ScaleControl, LocalTileLayer, GeoData
import pyrosm
from pyrosm import get_data
import rasterstats



# Import country coords
country_dict = get_country_centroids()
class Map(ipyleaflet.Map):

    def __init__(self, country: Optional[str] = None, **kwargs):
        self.bbox = None
        self.polygon_bounds = None

        super().__init__(**kwargs)
    
        if country:
            try:
                self.center = country_dict[country]['coords']
            except KeyError as err:
                print(f"KeyError: {err}. Please manually input center coordinates.")
            finally:
                self.country = country
        
        if 'zoom' not in kwargs:
            self.zoom = 11
        
        # Set default attributes
        if 'layout' not in kwargs:
            self.layout.height = "400px"
            self.layout.width = "600px"

        self.attribution_control = False

        # Add controls
        self.add_control(FullScreenControl())
        self.add_control(LayersControl(position='topright'))

        def handle_draw(target, action: str, geo_json: dict):
            print(action)
            print(geo_json)

        dc = (DrawControl(rectangle={'shapeOptions':{'color':"#a52a2a"}},
                                polyline = {"shapeOptions": {"color": "#6bc2e5", "weight": 2, "opacity": 1.0}},
                                polygon = {"shapeOptions": {"fillColor": "#eba134", "color": "#000000", "fillOpacity": 0.5, "weight":2},
                                            "drawError": {"color": "#dd253b", "message": "Delete and redraw"},
                                            "allowIntersection": False},
                                )
                    )
        dc.on_draw(handle_draw)
        self.add(dc)
        self.add_control(ScaleControl(position='bottomleft'))
        

        self.add_layer(basemap_to_tiles(basemaps.CartoDB.VoyagerNoLabels))
        self.add_layer(LocalTileLayer(path="http://mt1.google.com/vt/lyrs=m&x={x}&y={y}&z={z}", name = 'Google Streets'))
        self.add_layer(LocalTileLayer(path="http://mt1.google.com/vt/lyrs=s&x={x}&y={y}&z={z}", name = 'Google Hybrid'))
        self.add_layer(LocalTileLayer(path="http://mt1.google.com/vt/lyrs=p&x={x}&y={y}&z={z}", name = 'Google Terrain'))

    def add_bbox(self, show: bool = False, remove: bool = True):

        t_index = None
        
        # Find DrawControl layer index
        for i, control in enumerate(self.controls):
            if isinstance(control, DrawControl):
                t_index = i

        if self.controls[t_index].last_action == '':
            print("No bounding box/polygon found. Please draw on map.")
            if show == True:
                display(self)

        elif self.controls[t_index].last_action == 'created': 

            lon_list = []
            lat_list = []
            for lon,lat in [*self.controls[t_index].data[0]['geometry']['coordinates'][0]]:
                lon_list.append(lon)
                lat_list.append(lat)
        
            polygon_geom = Polygon(zip(lon_list, lat_list))
            gdf = gpd.GeoDataFrame(index=[0], geometry=[polygon_geom]) 

            # Assign bounding box as self object attribute
            self.polygon_bounds = gdf
            self.polygon_bounds.crs = 'epsg:4326'

            # Remove drawing on self and close display
            if remove == True: 
                print('Assigned bbox attribute to self object. Removing drawn boundary.')
                self.controls[t_index].clear()
            else: 
                print('Assigned bbox attribute to self object.')
    
    def add_polygon_boundary(
        self, 
        filepath: str,
        layer_name: Optional[str] = 'Site', 
        polygon_style: Optional[dict] = {'style': {'color': 'black', 'fillColor': '#3366cc', 'opacity':0.05, 'weight':1.9, 'dashArray':'2', 'fillOpacity':0.6},
                                         'hover_style': {'fillColor': 'red' , 'fillOpacity': 0.2}},                   
        show: bool = False, 
        remove: bool = True) -> None:

        if filepath:
            # GeoJSON string file
            gdf = gpd.read_file(filepath)
        
        # Assign polygon boundary attribute to polygon object
        self.polygon_bounds = gdf

        # Add polygon boundary as map layer
        geo_data = GeoData(geo_dataframe = gdf,
                   style=polygon_style['style'],
                   hover_style=polygon_style['hover_style'],
                   name = layer_name)
        self.add_layer(geo_data)
        if show:
            display(self)
        
    def remove_polygon_boundary(self) -> None:
        polygon_exists = False
        for i in self.layers:
            if isinstance(i, ipyleaflet.leaflet.GeoData):
                polygon_exists = True
        if polygon_exists:
            self.remove_layer(self.layers[len(self.layers)-1])
            print('Polygon bounding layer removed.')
        else:
            print('No polygon layer found on map.')

    def get_street_network(
            self, 
            directory: str,
            location: str,
            bandwidth: int = 200,
            network_type: str = 'driving',
            building_attr: bool = True,
            pop_attr: bool = True,
            poi_attr: bool = True,
            svi_attr: bool = False,
            use_tif: bool = False,
            layer_name: str = 'Street Network', 
            polyline_style: dict = {'style': {'color': 'black', 'opacity':0.5, 'weight':1.9},
                                'hover_style': {'color': 'yellow' , 'opacity': 0.2}},  
            show: bool = False, 
            dual: bool = False) -> MultiDiGraph:
        """Function to generate either primal planar or dual (edge) networks. Bandwidth allows to account for
        outside the network that share immediate neighbours with nodes within target area. *_attr arguments
        can be toggled on or off to allow computation of additional geographic information into networks.

        Args:
            directory (str): Location to store and download Geofabrik file. Download will be skipped if location file is found.
            location (str): Accepts city name or country name to obtain OpenStreetMap data.
            bandwidth (int): Distance to extract information beyond network. Defaults to 200.
            network_type (str): Specified OpenStreetMap transportation mode. Defaults to 'driving'.
            building_attr (bool): Specifies whether building morphology attributes should be included. Defaults to True.
            pop_attr (bool): Specifies whether population attributes should be included. Defaults to True.
            poi_attr (bool): Specifies whether points of interest attributes should be included. Defaults to True.
            use_tif (bool): Specifies whether csv or tif should be used to construct population attributes. Defaults to False.
            layer_name (str): Layer name to display in ipyleafet. Defaults to 'Street Network'.
            polyline_style (dict): Style dictionary for plotting vector layers in ipyleaflet. Defaults to {'style': {'color': 'black', 'opacity':0.5, 'weight':1.9}, 'hover_style': {'color': 'yellow' , 'opacity': 0.2}}.
            show (bool): Specifies whether a new map view should be created at line location. Defaults to False.
            dual (bool): If true, creates a dual (edge) network graph. Defaults to False.
            
        Raises:
            Exception: No bounding box or polygon found to construct network.

        Returns:
            MultiDiGraph: A networkx/osmnx primal planar or dual (edge) network with specified attribute information.
        """            

        start = time.time()
        try:
            fp = get_data(location, directory = directory)
            print('Creating data folder and downloading osm street data...')
        except KeyError:
            fp = get_data(self.country, directory = directory)
            print(f"KeyError: No pre-downloaded osm data available for specified city, will instead try for specified country.")
        finally:
            raise ValueError('No osm data found for specified location.')

        print('Data extracted successfully. Proceeding to construct street network.')

        # Project and buffer original polygon to examine nodes outside boundary
        try:
            original_bbox = self.polygon_bounds.geometry.values[0]
            buffered_tp = buffer_polygon(self.polygon_bounds, bandwidth=bandwidth)
            buffered_bbox = buffered_tp.geometry.values[0]
        # catch when it hasn't even been defined 
        except (AttributeError, NameError):
            raise Exception('Please delimit a bounding box.')

        # Obtain nodes and edges within buffered polygon
        osm = pyrosm.OSM(fp, bounding_box=buffered_bbox)
        nodes, edges = osm.get_network(network_type=network_type, nodes=True)

        if building_attr:
            # Get building spatial data and project 
            building = osm.get_buildings()
            building_proj = project_gdf(building)

            # Make geometry type homogeneous (polygons) to to allow overlay operation
            building_polygon = fill_and_expand(building_proj)

        # Build networkx graph for pre-processing
        G_buff = osm.to_graph(nodes, edges, graph_type="networkx", force_bidirectional=True, retain_all=True)
        
        # Add great circle length to network edges
        G_buff = add_edge_lengths(G_buff)

        # Simplify graph by removing nodes between endpoints and joining linestrings
        G_buff_simple = simplify_graph(G_buff)

        # Identify nodes inside and outside (buffered polygon) of original polygon
        gs_nodes = graph_to_gdf(G_buff_simple, nodes=True)[["geometry"]]
        to_keep = gs_nodes.within(original_bbox)
        to_keep = gs_nodes[to_keep]
        nodes_outside = gs_nodes[~gs_nodes.index.isin(to_keep.index)]
        set_outside = nodes_outside.index

        # Truncate network by edge if all neighbours fall outside original polygon
        nodes_to_remove = set()
        for node in set_outside:
            neighbors = set(G_buff_simple.successors(node)) | set(G_buff_simple.predecessors(node))
            if neighbors.issubset(nodes_outside):
                nodes_to_remove.add(node)
        
        G_buff_trunc = G_buff_simple.copy()
        initial = G_buff_trunc.number_of_nodes()
        G_buff_trunc.remove_nodes_from(nodes_to_remove)

        # Remove unconnected subgraphs
        max_wcc = max(nx.weakly_connected_components(G_buff_trunc), key=len)
        G_buff_trunc = nx.subgraph(G_buff_trunc, max_wcc)
        nodes, edges = graph_to_gdf(G_buff_trunc, nodes=True, edges=True)
        nodes = nodes.fillna('')

        # If not dual representation graph
        if dual == False:
            
            nodes_buffer = project_gdf(nodes)
            nodes_buffer['center_bound'] = nodes_buffer.geometry.buffer(bandwidth)
            nodes_buffer = nodes_buffer.set_geometry('center_bound')

            if building_attr:

                # Compute and add building attributes
                res_intersection = nodes_buffer.overlay(building_polygon, how='intersection')
                res_intersection['area'] = res_intersection.geometry.area
                area_series = res_intersection.groupby(['osmid'])['area'].sum()
                
                # Obtain proportion 
                total_area = math.pi*bandwidth**2
                area_series = area_series / total_area
                area_series.name = 'Footprint area'

                
                nodes = nodes.join(area_series)

                # Add perimeter
                res_intersection['perimeter'] = res_intersection.geometry.length
                perimeter_series = res_intersection.groupby(['osmid'])['perimeter'].sum()

                perimeter_series.name = 'Perimeter'
                nodes = nodes.join(perimeter_series)


                 # Add counts
                counts_series = res_intersection['osmid'].value_counts()
                counts_series.name = 'Counts'
                nodes = nodes.join(counts_series)
                
                nodes['Footprint area'] = nodes['Footprint area'].replace(np.nan, 0).astype(float)
                nodes['Perimeter'] = nodes['Perimeter'].replace(np.nan, 0).astype(float)
                nodes['Counts'] = nodes['Counts'].replace(np.nan, 0).astype(int)
            
            if pop_attr:

                groups = ['PopSum', 'Men', 'Women', 'Elderly','Youth','Children']

                if not use_tif:
                    pop_list, target_cols = get_population_data(self.country, use_tif=use_tif)
                    for i, data in enumerate(zip(pop_list, target_cols)):
                        proj_data = project_gdf(data[0])
                        res_intersection = proj_data.overlay(nodes_buffer, how='intersection')

                        # Add population sum
                        pop_sum_series = res_intersection.groupby(['osmid'])[data[1]].sum()
                        pop_sum_series.name = groups[i]
                        nodes = nodes.join(pop_sum_series)

                    for name in groups:
                        nodes[name] = nodes[name].replace(np.nan, 0).astype(float)
                    
                if use_tif: 
                    # Using .tif for computation
                    src_list, rc_list = get_population_data(self.country, use_tif=use_tif)
                    nodes_buffer = nodes_buffer.to_crs('epsg:4326')
                    
                    for group, src, rc in zip(groups, src_list, rc_list):
                        stats = ['sum']

                        result = rasterstats.zonal_stats(
                            nodes_buffer, 
                            rc[0], 
                            nodata = 0, 
                            affine = src.transform, 
                            stats = stats
                        )

                        result = pd.DataFrame(result)

                        # Add population sum
                        nodes[group] = list(result['sum'])
                        nodes[group] = nodes[group].replace(np.nan, 0).astype(int)
                
            if poi_attr:
                poi_path = pkg_resources.resource_filename('urbanity', "map_data/poi_filter.json")
                with open(poi_path) as poi_filter:
                    poi_filter = json.load(poi_filter)
                
                pois = osm.get_pois(custom_filter = poi_filter['custom_filter'])
                pois = pois.replace(np.nan, '')

                def poi_col(amenity, shop, tourism, leisure):
                    value = amenity
                    if amenity == '' and shop not in poi_filter['food_set']:
                        value = 'commercial'
                    elif amenity == '' and shop in poi_filter['food_set']:
                        value = shop
                    elif amenity == '' and tourism != '':
                        value = 'entertainment'
                    elif amenity == '' and leisure != '':
                        value = 'recreational'
                    return value
                
                pois['poi_col'] = pois.apply(lambda row: poi_col(row['amenity'], row['shop'], row['tourism'], row['leisure']), axis=1)
                pois = pois[['id', 'osm_type','lon','lat','name','poi_col','geometry']]

                pois = finetune_poi(pois, 'poi_col', poi_filter['replace_dict'], n=30)
                pois = project_gdf(pois)
                pois['geometry'] = pois.geometry.centroid

                res_intersection = pois.overlay(nodes_buffer, how='intersection')
                poi_series = res_intersection.groupby(['osmid'])['poi_col'].value_counts()
                pois_df = pd.DataFrame(index = poi_series.index, data = poi_series.values).reset_index()
                pois_df = pd.pivot(pois_df, index='osmid', columns='poi_col', values=0)
                nodes = nodes.merge(pois_df, how='left', left_index=True, right_index=True).replace(np.nan,0) 
                nodes = nodes.rename(columns = {'commercial':'Commercial', 'entertainment':'Entertainment','food':'Food','healthcare':'Healthcare','civic':'Civic', 'institutional':'Institutional'})

            if svi_attr:
                pass
                # svi_path = pkg_resources.resource_filename('urbanity', f"svi_data/{location}.geojson")
                # svi_data = gpd.read_file(svi_path)

                # # returns gdf of image id, tile_id, indicators, and point coords
                # res_intersection = svi_data.overlay(nodes_buffer, how='intersection')
                
                # indicators = ['Green View', 'Sky View', 'Building View', 'Road View']
                # for indicator in indicators:
                    
                #     svi_series = res_intersection.groupby(['osmid'])[indicator].mean()
                #     svi_series.name = indicator
                #     nodes = nodes.join(svi_series)

                #     index_mean = nodes[indicator].mean()
                #     nodes[indicator] = nodes[indicator].replace(np.nan, index_mean)
            
            print("--- %s seconds ---" % round(time.time() - start,3))
            return G_buff_trunc, nodes, edges

        elif dual: 
            # First extract dictionary of osmids and lengths for original nodes associated with each edge
            osmid_view = nx.get_edge_attributes(G_buff_trunc, "osmid")
            osmid_dict = {}
            for u,v in set(osmid_view.items()):
                if u not in osmid_dict:
                    osmid_dict[(u[:2])] = v
                else: 
                    osmid_dict[(u[:2])].append(v)

            length_view = nx.get_edge_attributes(G_buff_trunc, "length")
            length_dict = {}
            for u,v in set(length_view.items()):
                if u not in length_dict:
                    length_dict[(u[:2])] = v
                else: 
                    length_dict[(u[:2])].append(v)

            x_dict = nx.get_node_attributes(G_buff_trunc, "x")
            y_dict = nx.get_node_attributes(G_buff_trunc, "y")

            # Create new placeholder graph and add edges as nodes and adjacency links between edges as new edges
            L = nx.empty_graph(0)
            LG = nx.line_graph(G_buff_trunc)
            L.graph['crs'] = 'EPSG:4326'
            for node in set(G_buff_trunc.edges()):
                L.add_node(node, length = length_dict[node], osmids = osmid_dict[node], x = (x_dict[node[0]]+x_dict[node[1]])/2, y = (y_dict[node[0]]+y_dict[node[1]])/2, geometry=Point((x_dict[node[0]]+x_dict[node[1]])/2, (y_dict[node[0]]+y_dict[node[1]])/2))
            for u,v in set(LG.edges()):
                L.add_edge(u[:2],v[:2])

            # Extract nodes and edges GeoDataFrames from graph
            L_nodes, L_edges = graph_to_gdf(L, nodes=True, edges=True, dual=True)
            L_nodes = L_nodes.fillna('')

            if building_attr:
                building_nodes = project_gdf(L_nodes)
                building_nodes['center_bound'] = building_nodes.geometry.buffer(bandwidth)
                building_nodes = building_nodes.set_geometry('center_bound')
                building_nodes = building_nodes.reset_index()
                building_nodes = building_nodes.rename(columns = {'level_0':'from', 'level_1':'to'})
                building_nodes['unique'] = building_nodes.index

                # Compute and add area
                res_intersection = building_nodes.overlay(building_polygon, how='intersection')
                res_intersection['area'] = res_intersection.geometry.area
                area_series = res_intersection.groupby(['unique'])['area'].sum()
                
                # Obtain proportion 
                total_area = math.pi*bandwidth**2
                area_series = area_series / total_area
                area_series = area_series.astype(float)
                area_series.name = 'Footprint area'

                L_nodes['unique'] = list(range(len(L_nodes)))
                L_nodes = L_nodes.join(area_series, on = 'unique')

                # Add perimeter
                res_intersection['perimeter'] = res_intersection.geometry.length
                perimeter_series = res_intersection.groupby(['unique'])['perimeter'].sum()

                perimeter_series.name = 'Perimeter'
                L_nodes = L_nodes.join(perimeter_series, on = 'unique')


                 # Add counts
                counts_series = res_intersection['unique'].value_counts()
                counts_series.name = 'Counts'
                L_nodes = L_nodes.join(counts_series, on = 'unique')
                
                L_nodes['Footprint area'] = L_nodes['Footprint area'].replace(np.nan, 0).astype(float)
                L_nodes['Perimeter'] = L_nodes['Perimeter'].replace(np.nan, 0).astype(float)
                L_nodes['Counts'] = L_nodes['Counts'].replace(np.nan, 0).astype(int)
    

            print("--- %s seconds ---" % round(time.time() - start,3))

            return L, L_nodes, L_edges

    def remove_street_network(self) -> None:
        network_exists = False
        for i in self.layers:
            if isinstance(i, ipyleaflet.leaflet.GeoData):
                network_exists = True
        if network_exists:
            self.remove_layer(self.layers[len(self.layers)-1])
            print('Network layer removed.')
        else:
            print('No network layer found on map.')

    def get_building_network(
            self, 
            directory: str,
            layer_name: Optional[str] = 'Building Network', 
            polyline_style: Optional[dict] = {'style': {'color': 'black', 'opacity':0.5, 'weight':1.9},
                                'hover_style': {'color': 'yellow' , 'opacity': 0.2}},  
            show: Optional[bool] = False,
            network_type: str = 'driving') -> MultiDiGraph:
        start = time.time()
        print('Creating data folder and downloading osm building data...')
        fp = get_data(self.country, directory = directory)
        print('Data extracted successfully. Proceeding to construct building network.')

        # Project and buffer original polygon to examine nodes outside boundary
        try:
            original_bbox = self.polygon_bounds.geometry.values[0]
            buffered_tp = buffer_polygon(self.polygon_bounds, bandwidth=bandwidth)
            buffered_bbox = buffered_tp.geometry.values[0]
        # catch when it hasn't even been defined 
        except (AttributeError, NameError):
            raise Exception('Please delimit a bounding box.')

        # Obtain nodes and edges within buffered polygon
        osm = pyrosm.OSM(fp, bounding_box=buffered_bbox)

        # Get buildings
        buildings = osm.get_buildings()
        proj_buildings = project_gdf(buildings)

        # Get adjacency based on spatial intersection
        building_neighbours = {}
        for i,b in zip(proj_buildings.id,proj_buildings.geometry):
            s = proj_buildings.intersects(b.buffer(100))
            building_neighbours[i] = proj_buildings.id[s[s].index].values

        # Set centroid as geometry and create lon and lat columns
        proj_buildings['center'] = proj_buildings.geometry.centroid
        proj_buildings = proj_buildings.set_geometry("center")
        buildings = proj_buildings.to_crs(4326)
        buildings['x'] = buildings.geometry.x
        buildings['y'] = buildings.geometry.y
        buildings = buildings.set_index('id')

        # Create building network graph
        # Get dict of building nodes and their attributes
        id_to_attributes = {}
        for node in set(buildings.index):
            id_to_attributes[node] = buildings.loc[node].to_dict()

        B = nx.empty_graph(0)
        B.graph['crs'] = 'EPSG:4326'

        # Add nodes
        for node in set(buildings.index):
            B.add_node(node)
        nx.set_node_attributes(B, id_to_attributes)

        # Add edges
        for node, neighbours in building_neighbours.items():
            for neighbour in neighbours:
                B.add_edge(node, neighbour)

        # Compute euclidean distance between adjacent building centroids
        id_to_x = dict(zip(buildings.index, buildings.x))
        id_to_y = dict(zip(buildings.index, buildings.y))
        
        distance_between_buildings = {}
        for pair in set(B.edges()):
            distance_between_buildings[pair] = great_circle_vec(id_to_x[pair[0]], id_to_y[pair[0]], id_to_x[pair[1]], id_to_y[pair[1]])
        
        nx.set_edge_attributes(B, distance_between_buildings, 'length')

        # Identify largest weakly connected component
        max_wcc = max(nx.connected_components(B), key=len)
        B_max = nx.subgraph(B, max_wcc)

        B_nodes, B_edges = graph_to_gdf(B_max, nodes=True, edges=True, dual=True)
        B_nodes = B_nodes.fillna('')
        # Return graph and add network layer to Map object

        # building_data = GeoData(geo_dataframe = B_edges,
        #         style=polyline_style['style'],
        #         hover_style=polyline_style['hover_style'],
        #         name = layer_name)

        # self.add_layer(building_data)

        # if show:
        #     display(self)

        print("--- %s seconds ---" % round(time.time() - start,3))

        return B_max, B_nodes, B_edges


    



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['urbanity']

package_data = \
{'': ['*'], 'urbanity': ['map_data/*', 'svi_data/*']}

install_requires = \
['Shapely==1.8.4',
 'geopandas==0.11.1',
 'ipyleaflet==0.17.1',
 'networkx==2.8.6',
 'numpy==1.23.3',
 'pyrosm==0.6.1',
 'rasterio==1.3.2',
 'vt2geojson>=0.2.1,<0.3.0']

setup_kwargs = {
    'name': 'urbanity',
    'version': '0.3.0',
    'description': '',
    'long_description': None,
    'author': 'winstonyym',
    'author_email': 'winstonyym@u.nus.edu',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
